<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

namespace local_rwth_sd_sync;

defined('MOODLE_INTERNAL') || die();

use cache_helper;
use context_course;
use context_system;
use core_php_time_limit;
use core_tag_tag;
use moodle_exception;
use progress_trace;
use calendar_event;

use stdClass;

class sdapi {

    protected $config = null;
    private $verbose = false;

    public function __construct() {
        global $CFG;

        $verbose = get_config('local_rwth_sd_sync', 'verboselogging');
        $this->verbose = !empty($verbose);

        $debugging = get_config('local_rwth_sd_sync', 'debugging');
        if ($debugging) {
            $CFG->debug = DEBUG_DEVELOPER;
            $CFG->debugdisplay = 1;
        }
    }

    /**
     * Performs a course sync with external database.
     *
     * If necessary, it creates new courses or
     * updates course fields fullname and description
     *
     * @param progress_trace $trace
     * @return int 0 means success, 1 db connect failure, 4 db read failure
     */
    public function sync_courses(progress_trace $trace) {
        global $CFG, $DB;

        // Make sure we sync either enrolments or courses.
        if (!$this->get_config('dbtype') or !$this->get_config('newcoursetable') or !$this->get_config('newcoursefullname') or !$this->get_config('newcourseshortname')) {
            $trace->output('Course synchronisation skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Starting course synchronisation...');

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external enrolment database');
            $trace->finished();
            return 1;
        }

        $table           = trim($this->get_config('newcoursetable'));
        $fullname        = trim($this->get_config('newcoursefullname'));
        $shortname       = trim($this->get_config('newcourseshortname'));
        $idnumber        = trim($this->get_config('newcourseidnumber'));
        $category        = trim($this->get_config('newcoursecategory'));
        $summary         = trim($this->get_config('newcoursedescription'));
        $startdate       = trim($this->get_config('newcoursestartdate'));
        $enddate         = trim($this->get_config('newcourseenddate'));
        $customfieldsraw = trim($this->get_config('newcoursecustomfields'));

        // Parse custom course fields setting.
        $customfieldslines = preg_split("/(\r\n|\n|\r)/", $customfieldsraw);
        $customfields = [];           // Fields: customfield, value or (remotecolumn and remotecolumn_l), update.
        $customfieldcontrollers = []; // For use with \core_customfield\api functions. [customfieldid => controller]
        foreach ($customfieldslines as $line) {
            $field = new stdClass();

            if (empty($line)) {
                continue;
            }
            $values = explode(':', $line);
            if (count($values) != 2 && count($values) != 3) {
                $trace->output("Error while parsing plugin config setting 'newcoursecustomfields': non-empty line '$line' must contain two or three values separated by a colon");
                $trace->finished();
                return 5;
            }
            $customfieldshortname = trim($values[0]);
            if (!$DB->record_exists('customfield_field', ['shortname' => $customfieldshortname])) {
                $trace->output("Error while parsing plugin config setting 'newcoursecustomfields': shortname '$customfieldshortname' doesn't match any custom course field");
                $trace->finished();
                return 6;
            }
            $customfield = $DB->get_record('customfield_field', ['shortname' => $customfieldshortname]);
            $field->customfield = $customfield;

            // If righthand value is enclosed in double quotes it gets treated as a static value.
            // Backslashes and double quotes enclosed by the outer double quotes have to be escaped
            // by writing a backslash in front of them.
            $value = trim($values[1]);
            $matches = [];
            if (preg_match('/^"((?:[^"\\\\]|\\\\"|\\\\\\\\)*)"$/', $value, $matches)) {
                // Static value.
                $value = $matches[1];
                $value = str_replace('\\\\', '\\', $value);
                $value = str_replace('\\"', '"', $value);
                $value = $this->translate_custom_field_value($customfield, $value);
                $field->value = $value;
            } else {
                // Field in remote DB.
                $remotecolumn = $value;

                $field->remotecolumn = $remotecolumn;
                $field->remotecolumn_l = strtolower($remotecolumn);
            }

            $field->update = true;
            if (count($values) > 2) {
                $updatecustomfield = trim($values[2]);
                if ($updatecustomfield == 'update') {
                    $field->update = true;
                } else if ($updatecustomfield == 'no-update') {
                    $field->update = false;
                } else {
                    $trace->output("Error while parsing plugin config setting 'newcoursecustomfields': '$customfieldshortname' doesn't match any 'updatecustomfield' value (either 'update' or 'no-update')");
                    $trace->finished();
                    return 7;
                }
            }
            $customfields[] = $field;
            $customfieldcontrollers[$customfield->id] = \core_customfield\field_controller::create($customfield->id);
        }

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $fullname_l  = strtolower($fullname);
        $shortname_l = strtolower($shortname);
        $idnumber_l  = strtolower($idnumber);
        $category_l  = strtolower($category);
        $summary_l   = strtolower($summary);
        $startdate_l = strtolower($startdate);
        $enddate_l   = strtolower($enddate);

        $localcoursefield = $this->get_config('localcoursefield', 'idnumber');
        $localcategoryfield = $this->get_config('localcategoryfield', 'id');
        $defaultcategory    = $this->get_config('defaultcategory');

        // Determine field for identifying courses in external DB.
        switch($localcoursefield) {
            case 'idnumber':
                $coursefield_l = $idnumber_l;
                break;
            case 'shortname':
                $coursefield_l = $shortname_l;
                break;
            default:
                $extdb->Close();
                $trace->output('Invalid local course field configured: '.$coursefield_l);
                $trace->finished();
                return 5;
        }
        if (empty($coursefield_l)) {
            $extdb->Close();
            $trace->output('Remote value for local course field '.$coursefield_l.' empty');
            $trace->finished();
            return 6;
        }

        if (!$DB->record_exists('course_categories', ['id' => $defaultcategory])) {
            $trace->output("default course category does not exist!", 1);
            $categories = $DB->get_records('course_categories', [], 'sortorder', 'id', 0, 1);
            $first = reset($categories);
            $defaultcategory = $first->id;
        }

        $sqlfields = [$fullname, $shortname];
        if ($category) {
            $sqlfields[] = $category;
        }
        if ($idnumber) {
            $sqlfields[] = $idnumber;
        }
        if ($summary) {
            $sqlfields[] = $summary;
        }
        if ($startdate) {
            $sqlfields[] = $startdate;
        }
        if ($enddate) {
            $sqlfields[] = $enddate;
        }
        foreach ($customfields as $customfield) {
            if (isset($customfield->remotecolumn)) {
                $sqlfields[] = $customfield->remotecolumn;
            }
        }

        $sql = $this->db_get_sql($table, [], $sqlfields, true);
        $createcourses = [];
        $updatecourses = [];
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($fields = $rs->FetchRow()) {
                    $isUpdateCourse = false;
                    $fields = array_change_key_case($fields, CASE_LOWER);
                    $fields = $this->db_decode($fields);
                    if (empty($fields[$shortname_l]) || empty($fields[$fullname_l]) || empty($fields[$coursefield_l])) {
                        $trace->output("error: invalid external course record, shortname, fullname and $localcoursefield are mandatory: " . htmlentities(json_encode($fields)), 1);
                        continue;
                    }
                    if ($fields[$enddate_l] != 0 && $fields[$startdate_l] == 0) {
                        $defaultduration = get_config('moodlecourse', 'courseduration');
                        $trace->output('warning: course enddate provided without startdate for course '.$fields[$shortname_l].
                                ': setting startdate = (enddate - default course duration of '.$defaultduration.')', 1);
                        $fields[$startdate_l] = $fields[$enddate_l] - $defaultduration;
                    }
                    if ($DB->record_exists('course', [$localcoursefield => $fields[$coursefield_l]])) {
                        // Course exists but maybe needs an update.
                        $isUpdateCourse = true;
                        $updateCoursesTmp = $DB->get_records('course', [$localcoursefield => $fields[$coursefield_l]]);
                        if (count($updateCoursesTmp) > 1) {
                            $trace->output('warning: multiple courses with '.$localcoursefield.' '.$fields[$coursefield_l].
                                'found in Moodle, updating all of them', 1);
                        }
                    }

                    // Translate custom field values (e.g. for dropdown fields).
                    foreach ($customfields as $customfield) {
                        if (isset($customfield->remotecolumn_l)) {
                            $value = $fields[$customfield->remotecolumn_l];
                            $value = $this->translate_custom_field_value($customfield->customfield, $value);
                            $fields[$customfield->remotecolumn_l] = $value;
                        }
                    }

                    // If the course does not exist, create it
                    // else check if it needs an update
                    if (!$isUpdateCourse) {
                        $course = new \stdClass();
                        $course->fullname  = $fields[$fullname_l];
                        $course->shortname = $fields[$shortname_l];
                        $course->idnumber  = $idnumber ? $fields[$idnumber_l] : '';
                        $course->summary   = $summary ? $fields[$summary_l] : '';
                        $course->startdate = $startdate ? $fields[$startdate_l] : '';
                        $course->enddate   = $enddate ? $fields[$enddate_l] : '';
                        if ($category) {
                            if (empty($fields[$category_l])) {
                                // Empty category means use default.
                                $course->category = $defaultcategory;
                            } else if ($coursecategory = $DB->get_record('course_categories', array($localcategoryfield => $fields[$category_l]), 'id')) {
                                // Yay, correctly specified category!
                                $course->category = $coursecategory->id;
                                unset($coursecategory);
                            } else {
                                // Bad luck, better not continue because unwanted ppl might get access to course in different category.
                                $trace->output('error: invalid category '.$fields[$category_l].', can not create course: '.$localcoursefield.'='.$fields[$coursefield_l], 1);
                                continue;
                            }
                        } else {
                            $course->category = $defaultcategory;
                        }
                        $course->customfields = [];
                        foreach ($customfields as $customfield) {
                            if (isset($customfield->remotecolumn_l)) {
                                $value = $fields[$customfield->remotecolumn_l];
                            } else if (isset($customfield->value)) {
                                $value = $customfield->value;
                            } else {
                                // Should never happen.
                                $trace->output('warning: customfield '.$customfield->shortname.' has neither remote DB column nor static value, skipping', 1);
                                continue;
                            }
                            $course->customfields[$customfield->customfield->id] = $value;
                        }

                        $createcourses[] = $course;
                    } else {
                        foreach ($updateCoursesTmp as $updateCourse) {
                            $courseNeedsUpdate = false;
                            if ($updateCourse->fullname !== $fields[$fullname_l]) {
                                $trace->output("Updating course $updateCourse->shortname fullname from: $updateCourse->fullname to $fields[$fullname_l]");
                                $courseNeedsUpdate = true;
                                $updateCourse->fullname  = $fields[$fullname_l];
                            }

                            if ($summary and $updateCourse->summary !== $fields[$summary_l]) {
                                $trace->output("Updating course $updateCourse->shortname summary from: $updateCourse->summary to $fields[$summary_l]");
                                $courseNeedsUpdate = true;
                                $updateCourse->summary = $fields[$summary_l];
                            }

                            // Check if custom field value has changed.
                            $values = \core_customfield\api::get_instance_fields_data($customfieldcontrollers, $updateCourse->id);
                            foreach ($customfields as $customfield) {
                                if (!$customfield->update) {
                                    // Customfield value should get updated.
                                    continue;
                                }
                                if (isset($customfield->remotecolumn_l)) {
                                    $newvalue = $fields[$customfield->remotecolumn_l];
                                } else if (isset($customfield->value)) {
                                    $newvalue = $customfield->value;
                                } else {
                                    // Should never happen.
                                    $trace->output('warning: customfield '.$customfield->shortname.' has neither remote DB column nor static value, skipping', 1);
                                    continue;
                                }
                                $existingvalue = $values[$customfield->customfield->id]->get_value();
                                if ($newvalue != $existingvalue) {
                                    $courseNeedsUpdate = true;
                                    $updateCourse->customfields[$customfield->customfield->id] = $newvalue;
                                }
                            }

                            if ($courseNeedsUpdate) {
                                $updatecourses[] = $updateCourse;
                            }
                        }
                    }
                }
            }
            $rs->Close();
        } else {
            $extdb->Close();
            $trace->output('Error reading data from the external course table');
            $trace->finished();
            return 4;
        }
        unset($customfields);
        if ($createcourses) {
            require_once("$CFG->dirroot/course/lib.php");

            $templatecourse = $this->get_config('templatecourse');

            $template = false;
            if ($templatecourse) {
                if ($template = $DB->get_record('course', array('shortname' => $templatecourse))) {
                    $template = fullclone(course_get_format($template)->get_course());
                    if (!isset($template->numsections)) {
                        $template->numsections = course_get_format($template)->get_last_section_number();
                    }
                    unset($template->id);
                    unset($template->fullname);
                    unset($template->shortname);
                    unset($template->idnumber);
                    unset($template->summary);
                    unset($template->startdate);
                    unset($template->enddate);
                } else {
                    $trace->output("can not find template for new course!", 1);
                }
            }
            if (!$template) {
                $courseconfig = get_config('moodlecourse');
                $template = new \stdClass();
                //$template->summary        = '';
                $template->summaryformat  = FORMAT_HTML;
                $template->format         = $courseconfig->format;
                $template->numsections    = $courseconfig->numsections;
                $template->newsitems      = $courseconfig->newsitems;
                $template->showgrades     = $courseconfig->showgrades;
                $template->showreports    = $courseconfig->showreports;
                $template->maxbytes       = $courseconfig->maxbytes;
                $template->groupmode      = $courseconfig->groupmode;
                $template->groupmodeforce = $courseconfig->groupmodeforce;
                $template->visible        = $courseconfig->visible;
                $template->lang           = $courseconfig->lang;
                $template->enablecompletion = $courseconfig->enablecompletion;
                $template->groupmodeforce = $courseconfig->groupmodeforce;
                $template->startdate      = usergetmidnight(time());
                if ($courseconfig->courseenddateenabled) {
                    $template->enddate    = usergetmidnight(time()) + $courseconfig->courseduration;
                }
            }

            $courseidwriteback = $this->get_config('courseidwriteback');
            foreach ($createcourses as $fields) {
                $newcourse = clone($template);
                $newcourse->fullname  = $fields->fullname;
                $newcourse->shortname = $fields->shortname;
                $newcourse->idnumber  = $fields->idnumber;
                $newcourse->category  = $fields->category;
                $newcourse->summary   = $fields->summary;
                $newcourse->startdate = $fields->startdate;
                $newcourse->enddate   = $fields->enddate;

                // Detect duplicate data once again, above we can not find duplicates
                // in external data using DB collation rules...
                if (!empty($newcourse->idnumber) and $DB->record_exists('course', array('idnumber' => $newcourse->idnumber))) {
                    $trace->output("can not insert new course, duplicate idnumber detected: ".$newcourse->idnumber, 1);
                    continue;
                }
                // Only check for duplicate shortnames if we identify courses by their shortnames.
                if ($localcoursefield === 'shortname' && !empty($newcourse->shortname) && $DB->record_exists('course', array('shortname' => $newcourse->shortname))) {
                    $trace->output("can not insert new course, duplicate shortname detected: ".$newcourse->shortname, 1);
                    continue;
                }
                $trace->output("creating course: $newcourse->fullname, $newcourse->shortname, $newcourse->idnumber, $newcourse->category", 1);
                $c = $this->create_course($newcourse);
                $trace->output("created course with Moodle internal ID $c->id", 1);

                // Set custom course fields.
                if (!empty($fields->customfields)) {
                    $trace->output("setting custom fields for course with id $c->id: " . json_encode($fields->customfields), 1);
                    $customfields = \core_customfield\api::get_instance_fields_data($customfieldcontrollers, $c->id);
                    foreach ($fields->customfields as $customfieldid => $value) {
                        $customfield = $customfields[$customfieldid];
                        $customfield->set($customfield->datafield(), $value);
                        $customfield->set('value', $value ?? '');
                        $context = context_course::instance($c->id);
                        $customfield->set('contextid', $context->id);
                        $customfield->save();
                    }
                }

                if ($courseidwriteback) {
                    // Write Moodle internal course ID to SiteDirectory.
                    $moodleinternalid = trim($this->get_config('newcoursemoodleinternalid'));
                    $trace->output("Setting $moodleinternalid to $c->id ($c->idnumber) in SiteDirectory");
                    $extdb->Execute("UPDATE $table SET $moodleinternalid = $c->id WHERE $idnumber = '$c->idnumber'");
                }
            }

            unset($createcourses);
            unset($template);
        }

        if ($updatecourses) {
            require_once("$CFG->dirroot/course/lib.php");

            foreach ($updatecourses as $fields) {
                $trace->output("Updating course: $fields->shortname");

                // Update custom course fields.
                if (!empty($fields->customfields)) {
                    $trace->output("updating custom fields for course with id $fields->id: " . json_encode($fields->customfields), 1);
                    $customfields = \core_customfield\api::get_instance_fields_data($customfieldcontrollers, $fields->id);
                    foreach ($fields->customfields as $customfieldid => $value) {
                        $customfield = $customfields[$customfieldid];
                        $customfield->set($customfield->datafield(), $value);
                        $customfield->set('value', $value);
                        $context = context_course::instance($fields->id);
                        $customfield->set('contextid', $context->id);
                        $customfield->save();
                    }
                }
                unset($fields->customfields);

                // The update_course function also checks the start/end dates for validity. If for some reason
                // the existing Moodle course has invalid dates we don't want the entire sync to fail here.
                // So we just don't pass the dates to update_course().
                unset($fields->startdate);
                unset($fields->enddate);

                update_course($fields);
            }

            unset($updatecourses);
        }

        // Close db connection.
        $extdb->Close();

        $trace->output('...course synchronisation finished.');
        $trace->finished();

        return 0;
    }

    /**
     * Create a course and return a $course object
     *
     * Please note this functions does not verify any access control,
     * the calling code is responsible for all validation (usually it is the form definition).
     * 
     * This function has been copied from course/lib.php and the shortname
     * uniqueness requirement has been removed, i.e. with this function you can
     * create courses with shortnames that already exist in other courses.
     *
     * @param array $editoroptions course description editor options
     * @param object $data  - all the data needed for an entry in the 'course' table
     * @return object new course instance
     */
    protected function create_course($data, $editoroptions = NULL) {
        global $DB, $CFG;

        //check the categoryid - must be given for all new courses
        $category = $DB->get_record('course_categories', array('id'=>$data->category), '*', MUST_EXIST);

        // MODIFICATION START.
        // // Check if the shortname already exists.
        // if (!empty($data->shortname)) {
        //     if ($DB->record_exists('course', array('shortname' => $data->shortname))) {
        //         throw new moodle_exception('shortnametaken', '', '', $data->shortname);
        //     }
        // }
        // MODIFICATION END.

        // Check if the idnumber already exists.
        if (!empty($data->idnumber)) {
            if ($DB->record_exists('course', array('idnumber' => $data->idnumber))) {
                throw new moodle_exception('courseidnumbertaken', '', '', $data->idnumber);
            }
        }

        if ($errorcode = course_validate_dates((array)$data)) {
            throw new moodle_exception($errorcode);
        }

        // Check if timecreated is given.
        $data->timecreated  = !empty($data->timecreated) ? $data->timecreated : time();
        $data->timemodified = $data->timecreated;

        // place at beginning of any category
        $data->sortorder = 0;

        if ($editoroptions) {
            // summary text is updated later, we need context to store the files first
            $data->summary = '';
            $data->summary_format = FORMAT_HTML;
        }

        if (!isset($data->visible)) {
            // data not from form, add missing visibility info
            $data->visible = $category->visible;
        }
        $data->visibleold = $data->visible;

        $newcourseid = $DB->insert_record('course', $data);
        $context = context_course::instance($newcourseid, MUST_EXIST);

        if ($editoroptions) {
            // Save the files used in the summary editor and store
            $data = file_postupdate_standard_editor($data, 'summary', $editoroptions, $context, 'course', 'summary', 0);
            $DB->set_field('course', 'summary', $data->summary, array('id'=>$newcourseid));
            $DB->set_field('course', 'summaryformat', $data->summary_format, array('id'=>$newcourseid));
        }
        if ($overviewfilesoptions = course_overviewfiles_options($newcourseid)) {
            // Save the course overviewfiles
            $data = file_postupdate_standard_filemanager($data, 'overviewfiles', $overviewfilesoptions, $context, 'course', 'overviewfiles', 0);
        }

        // update course format options
        course_get_format($newcourseid)->update_course_format_options($data);

        $course = course_get_format($newcourseid)->get_course();

        fix_course_sortorder();
        // purge appropriate caches in case fix_course_sortorder() did not change anything
        cache_helper::purge_by_event('changesincourse');

        // new context created - better mark it as dirty
        $context->mark_dirty();

        // Trigger a course created event.
        $event = \core\event\course_created::create(array(
            'objectid' => $course->id,
            'context' => context_course::instance($course->id),
            'other' => array('shortname' => $course->shortname,
                'fullname' => $course->fullname)
        ));

        $event->trigger();

        // Setup the blocks
        blocks_add_default_course_blocks($course);

        // Create default section and initial sections if specified (unless they've already been created earlier).
        // We do not want to call course_create_sections_if_missing() because to avoid creating course cache.
        $numsections = isset($data->numsections) ? $data->numsections : 0;
        $existingsections = $DB->get_fieldset_sql('SELECT section from {course_sections} WHERE course = ?', [$newcourseid]);
        $newsections = array_diff(range(0, $numsections), $existingsections);
        foreach ($newsections as $sectionnum) {
            course_create_section($newcourseid, $sectionnum, true);
        }

        // Save any custom role names.
        save_local_role_names($course->id, (array)$data);

        // set up enrolments
        enrol_course_updated(true, $course, $data);

        // Update course tags.
        if (isset($data->tags)) {
            core_tag_tag::set_item_tags('core', 'course', $course->id, context_course::instance($course->id), $data->tags);
        }

        return $course;
    }

    private function translate_custom_field_value($customfield, $value) {
        if ($customfield->type == 'select') {
            $config = json_decode($customfield->configdata);
            $options = explode("\r\n", $config->options);
            if (!in_array($value, $options)) {
                $value = $config->defaultvalue;
            }
            return array_search($value, $options) + 1; // Options start at 1.
        }
        return $value;
    }

    /** appointment sync *****************************************************/

    private static function setUser($user = null) {
        global $CFG, $DB;

        if (is_object($user)) {
            $user = clone($user);
        } else if (!$user) {
            $user = new \stdClass();
            $user->id = 0;
            $user->mnethostid = $CFG->mnet_localhost_id;
        } else {
            $user = $DB->get_record('user', array('id' => $user));
        }
        unset($user->description);
        unset($user->access);
        unset($user->preference);

        // Ensure session is empty, as it may contain caches and user specific info.
        \core\session\manager::init_empty_session();

        \core\session\manager::set_user($user);
    }

    /**
     * Forces synchronisation of appointments with external database.
     * There is no room/location in moodle calendar events -> use description field
     *
     * @param progress_trace $trace
     * @return int 0 means success, 1 db connect failure, 2 db read failure
     */
    public function sync_appointments(progress_trace $trace) {
        global $DB;

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        // Check for entries in event sync table where the event
        // doesn't exist anymore in Moodle and remove them.
        $deletedevents = $DB->get_records_sql('
                              SELECT *
                              FROM {rwth_sd_sync_appointment} sa
                              WHERE NOT EXISTS (
                                  SELECT *
                                  FROM {event} e
                                  WHERE e.id = sa.moodle_eventid
                              )');
        foreach ($deletedevents as $deletedevent) {
            $params = [
                'context' => context_system::instance(),
                'other' => [
                    'moodle_eventid' => $deletedevent->moodle_eventid,
                    'sitedirectory_eventid' => $deletedevent->sitedirectory_eventid,
                ]];
            $event = event\synced_event_deleted::create($params);
            $event->trigger();

            $DB->execute('DELETE FROM {rwth_sd_sync_appointment} sa
                          WHERE sa.moodle_eventid = :meid',
                          ['meid' => $deletedevent->moodle_eventid]);
        }

        $ret = 0;

        $newcourseappointments = [];
        $obsoletecourseappointments = [];
        $updatedcourseappointments = [];
        $ret += 1 * $this->diff_course_appointments($trace, $newcourseappointments, $obsoletecourseappointments, $updatedcourseappointments);
        $newgroupappointments = [];
        $obsoletegroupappointments = [];
        $updatedgroupappointments = [];
        $ret += 10 * $this->diff_group_appointments($trace, $newgroupappointments, $obsoletegroupappointments, $updatedgroupappointments);
        $obsoleteappointments = array_merge($obsoletecourseappointments, $obsoletegroupappointments);

        $ret += 100 * $this->remove_appointments($trace, $obsoleteappointments);
        $ret += 1000 * $this->add_course_appointments($trace, $newcourseappointments);
        $ret += 10000 * $this->add_group_appointments($trace, $newgroupappointments);
        $ret += 100000 * $this->update_course_appointments($trace, $updatedcourseappointments);
        $ret += 1000000 * $this->update_group_appointments($trace, $updatedgroupappointments);
        return $ret;
    }

    private function diff_course_appointments(progress_trace $trace, &$newappointments, &$obsoleteappointments, &$updatedappointments) {
        global $DB;

        if (
            !$this->get_config('dbtype')
            || !$this->get_config('courseappointmenttable')
            || !$this->get_config('courseappointmentcoursefield')
            || !$this->get_config('courseappointmentroomfield')
            || !$this->get_config('courseappointmentlocationfield')
            || !$this->get_config('courseappointmentstartfield')
            || !$this->get_config('courseappointmentendfield')
            || !$this->get_config('courseappointmentkeywordfield')
            || !$this->get_config('courseappointmentronidfield')
            || !$this->get_config('courseappointmentsubjectfield')
        ) {
            $trace->output('Course appointment synchronisation skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Starting course appointment diff...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        $table         = $this->get_config('courseappointmenttable');
        $coursefield   = trim($this->get_config('courseappointmentcoursefield'));
        $roomfield     = trim($this->get_config('courseappointmentroomfield'));
        $locationfield = trim($this->get_config('courseappointmentlocationfield'));
        $subjectfield  = trim($this->get_config('courseappointmentsubjectfield'));
        $keywordfield  = trim($this->get_config('courseappointmentkeywordfield'));
        $startfield    = trim($this->get_config('courseappointmentstartfield'));
        $endfield      = trim($this->get_config('courseappointmentendfield'));
        $ronidfield    = trim($this->get_config('courseappointmentronidfield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $coursefield_l   = strtolower($coursefield); // "LV-Nummer"?
        $roomfield_l     = strtolower($roomfield); // Room number (Building|Room)
        $locationfield_l = strtolower($locationfield); // Name for the location (ie "ZuseLab" C3 instead of "1385|101")
        $subjectfield_l  = strtolower($subjectfield); // Description from RWTHonline ("TERMIN_BETREFF")
        $keywordfield_l  = strtolower($keywordfield); // Name of event
        $startfield_l    = strtolower($startfield); // Start of event
        $endfield_l      = strtolower($endfield); // End of event
        $ronidfield_l    = strtolower($ronidfield); // RWTHonline ID

        $localcoursefield = $this->get_config('localcoursefield', 'id');

        $sqlfields = array($coursefield, $roomfield, $locationfield, $subjectfield, $keywordfield, $startfield, $endfield, $ronidfield);

        // Get appointments from external database.
        $externalappointments = array();
        $sql = $this->db_get_sql($table, array(), $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($externalappointment = $rs->FetchRow()) {
                    $externalappointment = array_change_key_case($externalappointment, CASE_LOWER);
                    // calendar events without any time information do not make sense 
                    if (!$externalappointment[$startfield_l] || !$externalappointment[$endfield_l]) {
                        continue;
                    }
                    // event end before event start does not make sense
                    if ($externalappointment[$startfield_l] > $externalappointment[$endfield_l]) {
                        continue;
                    }
                    $ronid = $externalappointment[$ronidfield_l];
                    $externalappointments[$ronid] = $externalappointment;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $extdb->Close();
            return 2;
        }
        $extdb->Close();

        // Find all existing appointments with eventsource 'rwthonline'.
        $existingappointments = array();
        $sql = "SELECT e.id, e.name, e.description, e.timestart, e.timeduration, e.location, e.description, c.$localcoursefield, sa.sitedirectory_eventid
                      FROM {event} e
                      JOIN {rwth_sd_sync_appointment} sa ON (sa.eventsource = 'rwthonline' AND e.id = sa.moodle_eventid)
                      JOIN {course} c ON c.id = e.courseid
                      WHERE e.groupid = 0";
        $rs = $DB->get_recordset_sql($sql);
        foreach ($rs as $existingevent) {
            $ronid = $existingevent->sitedirectory_eventid;
            $existingappointments[$ronid] = $existingevent;
        }
        $rs->close();

        $obsoleteappointments = array_diff_key($existingappointments, $externalappointments);

        $commonappointments = array_intersect_key($existingappointments, $externalappointments);
        foreach ($commonappointments as $ronid => $existingrecord) {
            $externalrecord = $externalappointments[$ronid];
            // Check for termin name change.
            $existingname = $existingrecord->name;
            $externalname = $externalrecord[$keywordfield_l] ?? '';
            // Check for startdate change.
            date_default_timezone_set("Europe/Berlin");
            $existingstartdate = intval($existingrecord->timestart);
            $externalstartdate = strtotime($externalrecord[$startfield_l]);
            // Check for enddate change.
            $existingenddate = $existingrecord->timestart + $existingrecord->timeduration;
            $externalenddate = strtotime($externalrecord[$endfield_l]);
            // Check for room or location change.
            list($existingroom, $existinglocation) = self::parse_location_string($existingrecord->location ?? '');
            $externalroom = $externalrecord[$roomfield_l] ?? '';
            $externallocation = $externalrecord[$locationfield_l] ?? '';
            // Check for subject change.
            $existingsubject = $existingrecord->description;
            $externalsubject = $externalrecord[$subjectfield_l] ?? '';

            if ($existingname !== htmlentities($externalname) ||
                    $existingstartdate !== $externalstartdate ||
                    $existingenddate !== $externalenddate ||
                    $existingroom !== $externalroom ||
                    $existinglocation !== $externallocation ||
                    $existingsubject !== htmlentities($externalsubject)) {
                $now = time();
                if ($externalenddate > $now) {
                    $updatedappointments[$ronid] = $externalrecord;
                }
            }
        }

        $newappointments = array_diff_key($externalappointments, $existingappointments);
        foreach ($newappointments as $ronid => $newappointment) {
            $courseid = $DB->get_field('course', 'id', array($localcoursefield => $newappointment[$coursefield_l]));
            if (!$courseid) {
                $params = [
                    'context' => context_system::instance(),
                    'other' => [
                        'courselocalfield' => $localcoursefield,
                        'courselocalfieldvalue' => $newappointment[$coursefield_l],
                        'error' => 'The course doesn\'t exist in Moodle.'
                    ]];
                $event = event\course_event_synchronization_failed::create($params);
                $event->trigger();
                continue;
            }

            $newappointment['moodlecourseid'] = $courseid;
            $newappointments[$ronid] = $newappointment;
        }

        // we do not need these any more
        unset($externalappointments);
        unset($existingappointments);

        $trace->output('... course appointment diff finished.');
        $trace->finished();

        return 0;
    }

    private function diff_group_appointments(progress_trace $trace, &$newappointments, &$obsoleteappointments, &$updatedappointments) {
        global $DB;

        if (
            !$this->get_config('dbtype')
            || !$this->get_config('groupappointmenttable')
            || !$this->get_config('groupappointmentgroupidfield')
            || !$this->get_config('groupappointmentroomfield')
            || !$this->get_config('groupappointmentlocationfield')
            || !$this->get_config('groupappointmentstartfield')
            || !$this->get_config('groupappointmentendfield')
            || !$this->get_config('groupappointmentkeywordfield')
            || !$this->get_config('groupappointmentronidfield')
            || !$this->get_config('groupappointmentsubjectfield')
        ) {
            $trace->output('Group appointment synchronisation skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Starting group appointment diff...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        $table         = $this->get_config('groupappointmenttable');
        $groupidfield  = trim($this->get_config('groupappointmentgroupidfield'));
        $roomfield     = trim($this->get_config('groupappointmentroomfield'));
        $locationfield = trim($this->get_config('groupappointmentlocationfield'));
        $subjectfield  = trim($this->get_config('groupappointmentsubjectfield'));
        $keywordfield  = trim($this->get_config('groupappointmentkeywordfield'));
        $startfield    = trim($this->get_config('groupappointmentstartfield'));
        $endfield      = trim($this->get_config('groupappointmentendfield'));
        $ronidfield    = trim($this->get_config('groupappointmentronidfield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $groupidfield_l  = strtolower($groupidfield);
        $roomfield_l     = strtolower($roomfield);
        $locationfield_l = strtolower($locationfield);
        $subjectfield_l  = strtolower($subjectfield);
        $keywordfield_l  = strtolower($keywordfield);
        $startfield_l    = strtolower($startfield);
        $endfield_l      = strtolower($endfield);
        $ronidfield_l    = strtolower($ronidfield);

        $sqlfields = [$groupidfield, $roomfield, $locationfield, $subjectfield, $keywordfield, $startfield, $endfield, $ronidfield];

        // Get appointments from external database.
        $externalappointments = [];
        $sql = $this->db_get_sql($table, [], $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($externalappointment = $rs->FetchRow()) {
                    $externalappointment = array_change_key_case($externalappointment, CASE_LOWER);
                    // calendar events without any time information do not make sense 
                    if (!$externalappointment[$startfield_l] || !$externalappointment[$endfield_l]) {
                        continue;
                    }
                    // event end before event start does not make sense
                    if ($externalappointment[$startfield_l] > $externalappointment[$endfield_l]) {
                        continue;
                    }
                    $ronid = $externalappointment[$ronidfield_l];
                    $externalappointments[$ronid] = $externalappointment;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $extdb->Close();
            return 2;
        }

        // Close external db connection.
        $extdb->Close();

        // Find all existing group appointments with eventsource 'rwthonline'.
        $existingappointments = [];
        $sql = "SELECT e.id, e.name, e.description, e.timestart, e.timeduration, e.location, e.description, sg.sitedirectory_groupid, sa.sitedirectory_eventid
                      FROM {event} e
                      JOIN {rwth_sd_sync_appointment} sa ON (sa.eventsource = 'rwthonline' AND e.id = sa.moodle_eventid)
                      JOIN {groups} g ON g.id = e.groupid
                      LEFT JOIN {rwth_sd_sync_group} sg ON sg.moodle_groupid = g.id";
        $rs = $DB->get_recordset_sql($sql);
        foreach ($rs as $existingevent) {
            $ronid = $existingevent->sitedirectory_eventid;
            $existingappointments[$ronid] = $existingevent;
        }
        $rs->close();

        $obsoleteappointments = array_diff_key($existingappointments, $externalappointments);

        $commonappointments = array_intersect_key($existingappointments, $externalappointments);
        foreach ($commonappointments as $ronid => $existingrecord) {
            $externalrecord = $externalappointments[$ronid];
            // Check for termin name change.
            $existingname = $existingrecord->name;
            $externalname = $externalrecord[$keywordfield_l] ?? '';
            // Check for startdate change.
            date_default_timezone_set("Europe/Berlin");
            $existingstartdate = intval($existingrecord->timestart);
            $externalstartdate = strtotime($externalrecord[$startfield_l]);
            // Check for enddate change.
            $existingenddate = $existingrecord->timestart + $existingrecord->timeduration;
            $externalenddate = strtotime($externalrecord[$endfield_l]);
            // Check for room or location change.
            list($existingroom, $existinglocation) = self::parse_location_string($existingrecord->location ?? '');
            $externalroom = $externalrecord[$roomfield_l] ?? '';
            $externallocation = $externalrecord[$locationfield_l] ?? '';
            // Check for subject change.
            $existingsubject = $existingrecord->description;
            $externalsubject = $externalrecord[$subjectfield_l] ?? '';

            if ($existingname !== htmlentities($externalname) ||
                    $existingstartdate !== $externalstartdate ||
                    $existingenddate !== $externalenddate ||
                    $existingroom !== $externalroom ||
                    $existinglocation !== $externallocation ||
                    $existingsubject !== htmlentities($externalsubject)) {
                $now = time();
                if ($externalenddate > $now) {
                    $updatedappointments[$ronid] = $externalrecord;
                }
            }
        }

        $newappointments = array_diff_key($externalappointments, $existingappointments);
        foreach ($newappointments as $ronid => $newappointment) {
            $moodlegroup = $DB->get_record_sql("SELECT g.*
                                                    FROM {groups} g
                                                    JOIN {rwth_sd_sync_group} sg ON sg.moodle_groupid = g.id
                                                    WHERE sg.sitedirectory_groupid = ?",
                                                [$newappointment[$groupidfield_l]]);

            if (!$moodlegroup) {
                $params = [
                    //'courseid' => SITEID,
                    'context' => context_system::instance(),
                    'other' => [
                        'groupremotefield' => $groupidfield,
                        'groupremotefieldvalue' => $newappointment[$groupidfield_l],
                        'error' => 'The group doesn\'t exist in Moodle.'
                    ]];
                $event = event\group_event_synchronization_failed::create($params);
                $event->trigger();
                continue;
            }

            $newappointment['moodlecourseid'] = $moodlegroup->courseid;
            $newappointment['moodlegroupid'] = $moodlegroup->id;
            $newappointments[$ronid] = $newappointment;
        }

        // we do not need these any more
        unset($externalappointments);
        unset($existingappointments);

        $trace->output('... group appointment diff finished.');
        $trace->finished();

        return 0;
    }

    public static function make_location_string($room, $location) {
        if (!empty($room)) {
            $location = "[$room] $location";
        }

        return trim($location);
    }

    public static function parse_location_string($location) {
        $matches = [];
        if (preg_match('/^\\[(?P<room>.*?)\\] (?P<location>.*)$/', $location, $matches)) {
            return [$matches['room'], $matches['location']];
        } else if (preg_match('/^\\[(?P<room>.*?)\\]$/', $location, $matches)) {
            return [$matches['room'], ''];
        }
        return ['', $location];
    }

    private function add_course_appointments($trace, $newappointments) {
        global $CFG, $DB;
        if ($newappointments) {
            $newevents = array();
            require_once($CFG->dirroot . '/calendar/lib.php');

            $coursefield  = trim($this->get_config('courseappointmentcoursefield'));
            $roomfield    = trim($this->get_config('courseappointmentroomfield'));
            $locationfield= trim($this->get_config('courseappointmentlocationfield'));
            $subjectfield = trim($this->get_config('courseappointmentsubjectfield'));
            $keywordfield = trim($this->get_config('courseappointmentkeywordfield'));
            $startfield   = trim($this->get_config('courseappointmentstartfield'));
            $endfield     = trim($this->get_config('courseappointmentendfield'));
            $ronidfield   = trim($this->get_config('courseappointmentronidfield'));

            // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
            $coursefield_l  = strtolower($coursefield);
            $roomfield_l    = strtolower($roomfield);
            $locationfield_l= strtolower($locationfield);
            $subjectfield_l = strtolower($subjectfield);
            $keywordfield_l = strtolower($keywordfield);
            $startfield_l   = strtolower($startfield);
            $endfield_l     = strtolower($endfield);
            $ronidfield_l   = strtolower($ronidfield);

            $localcoursefield = $this->get_config('localcoursefield', 'id');

            $trace->output('Adding new course appointments ...');

            $transaction = $DB->start_delegated_transaction();
            $success = true;

            foreach ($newappointments as $ronid => $newappointment) {

                if (empty($newappointment['moodlecourseid'])) {
                    $trace->output('Skipping event for course with '.$localcoursefield.' \''.$newappointment[$coursefield_l].'\' because the course doesn\'t exist in Moodle ...');
                    continue;
                }

                if ($this->verbose) {
                    $trace->output('Adding event for course with '.$localcoursefield.' \''.$newappointment[$coursefield_l].'\' (id: \''.$newappointment['moodlecourseid'].'\'): '
                        .$newappointment[$startfield_l].', '
                        .$newappointment[$endfield_l].', '
                        .$newappointment[$ronidfield_l].', '
                        .$newappointment[$roomfield_l].', '
                        .$newappointment[$locationfield_l].', '
                        .$newappointment[$subjectfield_l].', '
                        .$newappointment[$keywordfield_l]);
                }

                date_default_timezone_set("Europe/Berlin");
                $starttime = strtotime($newappointment[$startfield_l]);
                $duration = strtotime($newappointment[$endfield_l]) - strtotime($newappointment[$startfield_l]);

                $newappointment[$subjectfield_l] = htmlentities($newappointment[$subjectfield_l]);
                $newappointment[$keywordfield_l] = htmlentities($newappointment[$keywordfield_l]);

                $newevent = [
                    'name' => $newappointment[$keywordfield_l],
                    'description' => $newappointment[$subjectfield_l],
                    'location' => self::make_location_string($newappointment[$roomfield_l], $newappointment[$locationfield_l]),
                    'courseid' => $newappointment['moodlecourseid'],
                    'component' => 'local_rwth_sd_sync',
                    'eventtype' => 'course',
                    'timestart' => $starttime,
                    'timeduration' => $duration];

                $result = array();
                $eventobj = new calendar_event($newevent);
                $result[] = $eventobj->create($newevent);

                // Insert eventsource entries.
                if ($result) {

                    foreach ($result as $createdevent) {
                        $eventsourceparams = array();
                        $eventsourceparams['eventsource'] = 'rwthonline';
                        $eventsourceparams['moodle_eventid'] = $createdevent->__get('id');
                        $eventsourceparams['sitedirectory_eventid'] = $ronid;

                        $success &= $DB->insert_record('rwth_sd_sync_appointment', $eventsourceparams, false);
                    }
                }
            }

            if ($success) {
                // Everything done smoothly, let's commit.
                $transaction->allow_commit();
            } else {
                $errormsg = 'Error inserting records into rwth_sd_sync_appointment';
                $transaction->rollback(new moodle_exception($errormsg));
                $trace->output($errormsg);
            }
            $transaction->dispose();

            unset($newevents);
            $trace->output('... finished adding new course appointments.');
        }
        return 0;
    }

    private function add_group_appointments($trace, $newappointments) {
        global $CFG, $DB;
        if ($newappointments) {
            $newevents = array();
            require_once("$CFG->dirroot/calendar/lib.php");

            $trace->output('Adding new group appointments ...');

            $groupidfield = trim($this->get_config('groupappointmentgroupidfield'));
            $roomfield    = trim($this->get_config('groupappointmentroomfield'));
            $locationfield= trim($this->get_config('groupappointmentlocationfield'));
            $subjectfield = trim($this->get_config('groupappointmentsubjectfield'));
            $keywordfield = trim($this->get_config('groupappointmentkeywordfield'));
            $startfield   = trim($this->get_config('groupappointmentstartfield'));
            $endfield     = trim($this->get_config('groupappointmentendfield'));
            $ronidfield   = trim($this->get_config('groupappointmentronidfield'));

            // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
            $groupidfield_l = strtolower($groupidfield);
            $roomfield_l    = strtolower($roomfield);
            $locationfield_l= strtolower($locationfield);
            $subjectfield_l = strtolower($subjectfield);
            $keywordfield_l = strtolower($keywordfield);
            $startfield_l   = strtolower($startfield);
            $endfield_l     = strtolower($endfield);
            $ronidfield_l   = strtolower($ronidfield);

            foreach ($newappointments as $ronid => $newappointment) {
                if (empty($newappointment['moodlecourseid']) || empty ($newappointment['moodlegroupid'])) {
                    $trace->output('Skipping event for group with '.$groupidfield.' \''.$newappointment[$groupidfield_l].'\' because the group doesn\'t exist in Moodle ...');
                    continue;
                }

                if ($this->verbose) {
                    $trace->output('Adding event for course with id \''.$newappointment['moodlecourseid'].'\' and group with id \''.$newappointment['moodlegroupid'].'\': '
                        .$newappointment[$startfield_l].', '
                        .$newappointment[$endfield_l].', '
                        .$newappointment[$ronidfield_l].', '
                        .$newappointment[$roomfield_l].', '
                        .$newappointment[$locationfield_l].', '
                        .$newappointment[$subjectfield_l].', '
                        .$newappointment[$keywordfield_l]);
                }

                date_default_timezone_set("Europe/Berlin");
                $starttime = strtotime($newappointment[$startfield_l]);
                $duration = strtotime($newappointment[$endfield_l]) - strtotime($newappointment[$startfield_l]);

                $newappointment[$subjectfield_l] = htmlentities($newappointment[$subjectfield_l]);
                $newappointment[$keywordfield_l] = htmlentities($newappointment[$keywordfield_l]);

                $newevent = [
                    'name' => $newappointment[$keywordfield_l],
                    'description' => $newappointment[$subjectfield_l],
                    'location' => self::make_location_string($newappointment[$roomfield_l], $newappointment[$locationfield_l]),
                    'courseid' => $newappointment['moodlecourseid'],
                    'groupid' => $newappointment['moodlegroupid'],
                    'component' => 'local_rwth_sd_sync',
                    'eventtype' => 'group',
                    'timestart' => $starttime,
                    'timeduration' => $duration];

                $result = array();
                $eventobj = new calendar_event($newevent);
                $result[] = $eventobj->create($newevent);

                // Insert eventsource entries.
                if ($result) {

                    $transaction = $DB->start_delegated_transaction();
                    $success = true;

                    foreach ($result as $createdevent) {
                        $eventsourceparams = array();
                        $eventsourceparams['eventsource'] = 'rwthonline';
                        $eventsourceparams['moodle_eventid'] = $createdevent->__get('id');
                        $eventsourceparams['sitedirectory_eventid'] = $ronid;

                        $success &= $DB->insert_record('rwth_sd_sync_appointment', $eventsourceparams, false);
                    }

                    if ($success) {
                        // Everything done smoothly, let's commit.
                        $transaction->allow_commit();
                    } else {
                        $errormsg = 'Error inserting records into rwth_sd_sync_appointment';
                        $transaction->rollback(new moodle_exception($errormsg));
                        $trace->output($errormsg);
                    }
                    $transaction->dispose();
                }
            }

            unset($newevents);
            $trace->output('... finished adding new group appointments.');
        }
        return 0;
    }

    private function remove_appointments($trace, $obsoleteappointments) {
        global $CFG, $DB, $USER;
        if ($obsoleteappointments) {
            $obsoleteeventids = [];
            require_once("$CFG->dirroot/calendar/externallib.php");

            $trace->output('Removing obsolete appointments ...');

            // Remove eventsource entries.
            $transaction = $DB->start_delegated_transaction();
            $success = true;

            foreach ($obsoleteappointments as $obsoleteappointment) {

                if ($this->verbose) {
                    $trace->output('Event to be deleted: moodleid '.$obsoleteappointment->id);
                }

                // Remove eventsource entries.
                $success &= $DB->delete_records('rwth_sd_sync_appointment', array('moodle_eventid' => $obsoleteappointment->id));

                $obsoleteeventids[] = $obsoleteappointment->id;
            }

            if ($success) {
                // Everything done smoothly, let's commit.
                $transaction->allow_commit();
            } else {
                $errormsg = 'Error deleting records from rwth_sd_sync_appointment';
                $transaction->rollback(new moodle_exception($errormsg));
                $trace->output($errormsg);
            }
            $transaction->dispose();

            self::delete_calendar_events($obsoleteeventids);

            $trace->output('... finished removing obsolete appointments.');
        }
        return 0;
    }

    private function update_course_appointments($trace, $updatedappointments) {
        global $CFG, $DB;
        if (!$updatedappointments) {
            return 0;
        }
        require_once($CFG->dirroot . '/calendar/lib.php');

        $trace->output('Starting course appointment update...');

        $roomfield     = trim($this->get_config('courseappointmentroomfield'));
        $locationfield = trim($this->get_config('courseappointmentlocationfield'));
        $subjectfield  = trim($this->get_config('courseappointmentsubjectfield'));
        $keywordfield  = trim($this->get_config('courseappointmentkeywordfield'));
        $startfield    = trim($this->get_config('courseappointmentstartfield'));
        $endfield      = trim($this->get_config('courseappointmentendfield'));
        $ronidfield    = trim($this->get_config('courseappointmentronidfield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $roomfield_l     = strtolower($roomfield);
        $locationfield_l = strtolower($locationfield);
        $subjectfield_l  = strtolower($subjectfield);
        $keywordfield_l  = strtolower($keywordfield);
        $startfield_l    = strtolower($startfield);
        $endfield_l      = strtolower($endfield);
        $ronidfield_l    = strtolower($ronidfield);

        foreach ($updatedappointments as $appointment) {
            $sitedireventid = $appointment[$ronidfield_l];
            $moodleeventid = $DB->get_field('rwth_sd_sync_appointment', 'moodle_eventid', ['sitedirectory_eventid' => $sitedireventid]);
            if (!$moodleeventid) {
                $trace->output('Warning: Course event with Sitedirectory ID '.$sitedireventid.' not found in mapping table. Cannot find corresponding Moodle event. Therefore not updating the Moodle event. This should not happen!');
                continue;
            }
            $appointmentdata = new stdClass();
            $appointmentdata->id = $moodleeventid;
            $appointmentdata->name = htmlentities($appointment[$keywordfield_l] ?? '');
            $appointmentdata->description = htmlentities($appointment[$subjectfield_l] ?? '');
            $eventroom = $appointment[$roomfield_l] ?? '';
            $eventlocation = $appointment[$locationfield_l] ?? '';
            $appointmentdata->location = self::make_location_string($eventroom, $eventlocation);
            date_default_timezone_set("Europe/Berlin");
            $starttime = strtotime($appointment[$startfield_l]);
            $duration = strtotime($appointment[$endfield_l]) - strtotime($appointment[$startfield_l]);
            $appointmentdata->timestart = $starttime;
            $appointmentdata->timeduration = $duration;
            $appointmentdata->component = 'local_rwth_sd_sync';
            $appointmentdata->eventtype = 'course';
            $eventobj = new calendar_event($appointmentdata);
            $eventobj->update($appointmentdata, false);
        }
        $trace->output('... course appointment update finished.');
        $trace->finished();
        return 0;
    }

    private function update_group_appointments($trace, $updatedappointments) {
        global $CFG, $DB;
        if (!$updatedappointments) {
            return 0;
        }
        require_once($CFG->dirroot . '/calendar/lib.php');

        $trace->output('Starting group appointment update...');

        $roomfield     = trim($this->get_config('groupappointmentroomfield'));
        $locationfield = trim($this->get_config('groupappointmentlocationfield'));
        $subjectfield  = trim($this->get_config('groupappointmentsubjectfield'));
        $keywordfield  = trim($this->get_config('groupappointmentkeywordfield'));
        $startfield    = trim($this->get_config('groupappointmentstartfield'));
        $endfield      = trim($this->get_config('groupappointmentendfield'));
        $ronidfield    = trim($this->get_config('groupappointmentronidfield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $roomfield_l     = strtolower($roomfield);
        $locationfield_l = strtolower($locationfield);
        $subjectfield_l  = strtolower($subjectfield);
        $keywordfield_l  = strtolower($keywordfield);
        $startfield_l    = strtolower($startfield);
        $endfield_l      = strtolower($endfield);
        $ronidfield_l    = strtolower($ronidfield);

        foreach ($updatedappointments as $appointment) {
            $sitedireventid = $appointment[$ronidfield_l];
            $moodleeventid = $DB->get_field('rwth_sd_sync_appointment', 'moodle_eventid', ['sitedirectory_eventid' => $sitedireventid]);
            if (!$moodleeventid) {
                $trace->output('Warning: Group event with Sitedirectory ID '.$sitedireventid.' not found in mapping table. Cannot find corresponding Moodle event. Therefore not updating the Moodle event. This should not happen!');
                continue;
            }
            $appointmentdata = new stdClass();
            $appointmentdata->id = $moodleeventid;
            $appointmentdata->name = htmlentities($appointment[$keywordfield_l] ?? '');
            $appointmentdata->description = htmlentities($appointment[$subjectfield_l] ?? '');
            $eventroom = $appointment[$roomfield_l] ?? '';
            $eventlocation = $appointment[$locationfield_l] ?? '';
            $appointmentdata->location = self::make_location_string($eventroom, $eventlocation);
            date_default_timezone_set("Europe/Berlin");
            $starttime = strtotime($appointment[$startfield_l]);
            $duration = strtotime($appointment[$endfield_l]) - strtotime($appointment[$startfield_l]);
            $appointmentdata->timestart = $starttime;
            $appointmentdata->timeduration = $duration;
            $appointmentdata->component = 'local_rwth_sd_sync';
            $appointmentdata->eventtype = 'group';
            $eventobj = new calendar_event($appointmentdata);
            $eventobj->update($appointmentdata, false);
        }
        $trace->output('... group appointment update finished.');
        $trace->finished();
        return 0;
    }

    /** appointment sync end **************************************************/

    /** group sync ************************************************************/

    public function sync_groups(progress_trace $trace) {
        global $CFG, $DB;

        require_once("$CFG->dirroot/group/lib.php");

        if (!$this->get_config('dbtype')
            || !$this->get_config('grouptable')
            || !$this->get_config('groupcoursefield')
            || !$this->get_config('groupidfield')
            || !$this->get_config('groupnamefield')
        ) {
            $trace->output('group synchronisation skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Starting group synchronisation...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        $table           = $this->get_config('grouptable');
        $course_field    = trim($this->get_config('groupcoursefield'));
        $id_field        = trim($this->get_config('groupidfield'));
        $name_field      = trim($this->get_config('groupnamefield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $course_field_l    = strtolower($course_field);
        $id_field_l        = strtolower($id_field);
        $name_field_l      = strtolower($name_field);

        $localcoursefield = $this->get_config('localcoursefield', 'id');
        $localcoursefieldsql = "course_$localcoursefield";

        $sqlfields = [
            $course_field,
            $id_field,
            $name_field
        ];

        // Get groups from external database.
        $externalgroups = [];
        $sql = $this->db_get_sql($table, array(), $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($externalgroup = $rs->FetchRow()) {
                    $externalgroup = array_change_key_case($externalgroup, CASE_LOWER);
                    $sitedirectory_groupid = $externalgroup[$id_field_l];
                    if (empty($sitedirectory_groupid)) {
                        $trace->output('Skipping group with empty '.$id_field_l);
                        continue;
                    }
                    $externalgroups[$sitedirectory_groupid] = (object)$externalgroup;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $trace->finished();
            $extdb->Close();
            return 2;
        }

        // Close external db connection.
        $extdb->Close();

        // Find all existing groups synced with this plugin.
        $existinggroups = [];
        $sql = "SELECT
                    m.id,
                    m.courseid,
                    m.name,
                    s.sitedirectory_groupid,
                    c.$localcoursefield AS $localcoursefieldsql
                FROM {groups} m
                JOIN {rwth_sd_sync_group} s ON (m.id = s.moodle_groupid)
                JOIN {course} c ON c.id = m.courseid";
        $rs = $DB->get_recordset_sql($sql);
        foreach ($rs as $existinggroup) {
            $sitedirectory_groupid = $existinggroup->sitedirectory_groupid;
            $existinggroups[$sitedirectory_groupid] = $existinggroup;
        }
        $rs->close();

        // Delete all entries in rwth_sd_sync_group for groups that were manually deleted in moodle.
        $sql = "DELETE s.*
                      FROM {rwth_sd_sync_group} s
                      WHERE NOT EXISTS(
                          SELECT g.id
                          FROM {groups} g
                          WHERE g.id = s.moodle_groupid
                      )";
        $DB->execute($sql);

        // Create groups.
        $newgroups = array_diff_key($externalgroups, $existinggroups);
        foreach ($newgroups as $newgroup) {
            $courseid = $DB->get_field('course', 'id', [$localcoursefield => $newgroup->$course_field_l]);
            if (!$courseid) {
                $trace->output('Skipping creation of group with '.$id_field.' '.$newgroup->$id_field_l
                        .' because course with '.$localcoursefield.' '.$newgroup->$course_field_l
                        .' doesn\'t exist');
                continue;
            }

            $data = new stdClass();
            $data->courseid = $courseid;
            $data->name = $newgroup->$name_field_l;
            $trace->output('Creating group with '.$id_field.' '.$newgroup->$id_field_l.' in course with '
                    .$localcoursefield.' '.$newgroup->$course_field_l);
            $newgroupid = groups_create_group($data);
            if (!$newgroupid) {
                $trace->output('Error creating group');
                $trace->finished();
                return 3;
            }

            try {
                $data = new stdClass();
                $data->sitedirectory_groupid = $newgroup->$id_field_l;
                $data->moodle_groupid = $newgroupid;
                $DB->insert_record('rwth_sd_sync_group', $data);
            } catch (moodle_exception $e) {
                // Make sure the DB is in a consistent state.
                groups_delete_group($newgroupid);
                throw $e;
            }
        }

        // Update groups.
        $updatablegroups = array_intersect_key($existinggroups, $externalgroups);
        foreach ($updatablegroups as $sitedirectory_groupid => $updatablegroup) {
            $externalgroup = $externalgroups[$sitedirectory_groupid];
            if ($updatablegroup->$localcoursefieldsql != $externalgroup->$course_field_l
                    || $updatablegroup->name != $externalgroup->$name_field_l) {
                $courseid = $DB->get_field('course', 'id', [$localcoursefield => $externalgroup->$course_field_l]);
                if (!$courseid) {
                    $trace->output('Skipping update of group with '.$id_field.' '.$externalgroup->$id_field_l
                            .' because course with '.$localcoursefield.' '.$externalgroup->$course_field_l
                            .' doesn\'t exist');
                    continue;
                }
                $data = new stdClass();
                $data->id = $updatablegroup->id;
                $data->courseid = $courseid;
                $data->name = $externalgroup->$name_field_l;
                $trace->output('Updating group with '.$id_field.' '.$externalgroup->$id_field_l
                        .' in course with '.$localcoursefield.' '.$externalgroup->$course_field_l);
                try {
                    groups_update_group($data);
                } catch (moodle_exception $e) {
                    $trace->output('Error updating group: '.$e->getMessage());
                    $trace->finished();
                    return 4;
                }
            }
        }

        // Delete groups.
        $obsoletegroups = array_diff_key($existinggroups, $externalgroups);
        foreach ($obsoletegroups as $obsoletegroup) {
            $DB->delete_records('rwth_sd_sync_group', [
                'moodle_groupid' => $obsoletegroup->id
            ]);

            $trace->output('Deleting group with '.$id_field.' '.$obsoletegroup->sitedirectory_groupid
                    .' from course with '.$localcoursefield.' '.$obsoletegroup->$localcoursefieldsql);
            if (!groups_delete_group($obsoletegroup->id)) {
                $trace->output('Error deleting course');
                $trace->finished();
                return 5;
            }
        }
        $trace->output('... group synchronisation finished.');
        $trace->finished();

        return 0;
    }

    public function sync_group_participants(progress_trace $trace) {
        global $CFG, $DB;

        require_once("$CFG->dirroot/group/lib.php");

        if (!$this->get_config('dbtype')
            || !$this->get_config('groupparticipantstable')
            || !$this->get_config('groupparticipantsgroupidfield')
            || !$this->get_config('groupparticipantsusernamefield')
        ) {
            $trace->output('group participants synchronisation skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Starting group participants synchronisation...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        $table           = $this->get_config('groupparticipantstable');
        $groupid_field   = trim($this->get_config('groupparticipantsgroupidfield'));
        $username_field  = trim($this->get_config('groupparticipantsusernamefield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $groupid_field_l   = strtolower($groupid_field);
        $username_field_l  = strtolower($username_field);

        $sqlfields = [
            $groupid_field,
            $username_field
        ];

        // Get group participants from external database.
        $externalpartsraw = [];
        $sql = $this->db_get_sql($table, array(), $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($part = $rs->FetchRow()) {
                    $part = array_change_key_case($part, CASE_LOWER);
                    $externalpartsraw[] = (object)$part;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $trace->finished();
            $extdb->Close();
            return 2;
        }

        // Close external db connection.
        $extdb->Close();

        // Translate RON groupid to Moodle groupid and Moodle username to Moodle userid.
        $externalparts = [];
        foreach ($externalpartsraw as $rawpart) {
            $sitedirectory_groupid = $rawpart->$groupid_field_l;
            $username = $rawpart->$username_field_l;

            $moodle_groupid = $DB->get_field('rwth_sd_sync_group', 'moodle_groupid', ['sitedirectory_groupid' => $sitedirectory_groupid]);
            if (!$moodle_groupid) {
                $trace->output('Skipping participation of user with '.$username_field.' '.$username
                        .' in group with '.$groupid_field.' '.$sitedirectory_groupid.' because group'
                        .' doesn\'t exist');
                continue;
            }
            $userid = $DB->get_field('user', 'id', ['username' => $username]);
            if (!$userid) {
                $trace->output('Skipping participation of user with '.$username_field.' '.$username
                        .' in group with '.$groupid_field.' '.$sitedirectory_groupid.' because user'
                        .' doesn\'t exist');
                continue;
            }
            $part = new stdClass();
            $part->groupid = $moodle_groupid;
            $part->userid = $userid;
            $externalparts[] = $part;
        }

        $lastsyncparts = [];
        foreach ($DB->get_records('rwth_sd_sync_group_member') as $record) {
            $groupid = $record->moodle_groupid;
            $userid = $record->moodle_userid;

            if (!$DB->record_exists('groups_members', ['groupid' => $groupid, 'userid' => $userid])) {
                // User has been removed from group directly in moodle, not via sitedirectory.
                // => Remove entry from group member sync table.
                $cond = [
                    'moodle_groupid' => $groupid,
                    'moodle_userid' => $userid
                ];
                $DB->delete_records('rwth_sd_sync_group_member', $cond);
                continue;
            }

            $part = new stdClass();
            $part->groupid = $groupid;
            $part->userid = $userid;
            $lastsyncparts[] = $part;
        }

        $partscompare = function($a, $b) {
            if ($a->groupid != $b->groupid) {
                return $a->groupid - $b->groupid;
            }
            if ($a->userid != $b->userid) {
                return $a->userid - $b->userid;
            }
            return 0;
        };

        // Remove users from groups.
        $obsoleteparts = array_udiff($lastsyncparts, $externalparts, $partscompare);
        foreach ($obsoleteparts as $obsoletepart) {
            $groupid = $obsoletepart->groupid;
            $userid = $obsoletepart->userid;
            $trace->output('Removing user with id '.$userid.' from group with id '
                    .$groupid.' ...');
            if (!groups_remove_member($groupid, $userid)) {
                $trace->output('Could not remove user from group');
                continue;
            }
            $cond = [
                'moodle_groupid' => $groupid,
                'moodle_userid' => $userid
            ];
            if ($DB->record_exists('rwth_sd_sync_group_member', $cond)) {
                $DB->delete_records('rwth_sd_sync_group_member', $cond);
            }
        }

        // Add new users to groups.
        $newparts = array_udiff($externalparts, $lastsyncparts, $partscompare);
        foreach ($newparts as $newpart) {
            $groupid = $newpart->groupid;
            $userid = $newpart->userid;
            $trace->output('Adding user with id '.$userid.' to group with id '
                    .$groupid.' ...');
            if (!groups_add_member($groupid, $userid)) {
                $trace->output('Could not add user to group');
                continue;
            }
            $cond = [
                'moodle_groupid' => $groupid,
                'moodle_userid' => $userid
            ];
            if (!$DB->record_exists('rwth_sd_sync_group_member', $cond)) {
                $DB->insert_record('rwth_sd_sync_group_member', (object)$cond);
            }
        }

        $trace->output('... group participants synchronisation finished.');
        $trace->finished();

        return 0;
    }

    /** group sync end ********************************************************/

    /** upgrade functions *****************************************************/

    public function write_courseids_to_sitedirectory(progress_trace $trace) {
        global $DB;

        if (!$this->get_config('courseidwriteback')
            || !$this->get_config('dbtype')
            || !$this->get_config('newcoursetable')
            || !$this->get_config('newcourseidnumber')
        ) {
            $trace->output('Writing Moodle internal course IDs to SiteDirectory skipped.');
            $trace->finished();
            return 0;
        }

        $trace->output('Start writing course IDs to Sitedirectory...');

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with SiteDirectory');
            $trace->finished();
            return 1;
        }

        $remotecoursetable = $this->get_config('newcoursetable');
        $remotecourseidnumberfield = trim($this->get_config('newcourseidnumber'));
        $remotemoodleinternalidfield = trim($this->get_config('newcoursemoodleinternalid'));
        $localcoursefield = $this->get_config('localcoursefield');

        $remotecourseidnumberfield_l = strtolower($remotecourseidnumberfield);

        $sql = $this->db_get_sql($remotecoursetable, [], ['id', $remotecourseidnumberfield], true);
        $rs = $extdb->Execute($sql);
        if (!$rs) {
            $trace->output('Error getting content of remote course table');
            $trace->finished();
            return 2;
        }
        while ($fields = $rs->FetchRow()) {
            $fields = array_change_key_case($fields, CASE_LOWER);
            $fields = $this->db_decode($fields);

            $remoteidnumber = $fields[$remotecourseidnumberfield_l];
            $id = $DB->get_field('course', 'id', [$localcoursefield => $remoteidnumber]);
            if (!$id) {
                $trace->output("Skipping course with $localcoursefield $remoteidnumber, because it doesn't exist in Moodle");
            }

            // Write Moodle internal course ID to SiteDirectory.
            $trace->output("Setting $remotemoodleinternalidfield to $id ($remoteidnumber) in SiteDirectory");
            $extdb->Execute("UPDATE $remotecoursetable SET $remotemoodleinternalidfield = $id WHERE $remotecourseidnumberfield = '$remoteidnumber'");
        }

        $trace->output('...finished writing course IDs to Sitedirectory.');
        $trace->finished();
        return 0;
    }

    /**
     * Only used during upgrades.
     * 
     * In old versions of the DB we didn't store the sitedirectory id of an appointment. This function
     * is used during upgrade so that we can then set a unique constrint on the sitedirectory id afterwards.
     */
    public function add_sitedirectory_ids_for_appointments(progress_trace $trace) {
        $ret = 0;
        // $ret +=   1 * $this->add_sitedirectory_ids_for_course_appointments($trace);
        $ret += 100 * $this->add_sitedirectory_ids_for_group_appointments($trace);
        return $ret;
    }

    private function add_sitedirectory_ids_for_course_appointments(progress_trace $trace) {
        global $CFG, $DB, $USER;

        if (
            !$this->get_config('dbtype')
            || !$this->get_config('courseappointmenttable')
            || !$this->get_config('courseappointmentcoursefield')
            || !$this->get_config('courseappointmentkeywordfield')
            || !$this->get_config('courseappointmentstartfield')
            || !$this->get_config('courseappointmentendfield')
        ) {
            $syncedappointmentsexist = $DB->record_exists('rwth_sd_sync_appointment', []);
            if ($syncedappointmentsexist) {
                $trace->output('Cannot get sitedirectory ids for synced appointments because config settings are missing.');
                $trace->finished();
                return 3;
            } else {
                $trace->output('No synced appointments exist. Nothing to do.');
                $trace->finished();
                return 0;
            }
        }

        $trace->output('Starting update of sitedirectory ids for course appointments ...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        $table        = $this->get_config('courseappointmenttable');
        $coursefield  = trim($this->get_config('courseappointmentcoursefield'));
        $roomfield    = trim($this->get_config('courseappointmentroomfield'));
        $subjectfield = trim($this->get_config('courseappointmentsubjectfield'));
        $keywordfield = trim($this->get_config('courseappointmentkeywordfield'));
        $startfield   = trim($this->get_config('courseappointmentstartfield'));
        $endfield     = trim($this->get_config('courseappointmentendfield'));
        $ronidfield   = trim($this->get_config('courseappointmentronidfield'));

        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $coursefield_l  = strtolower($coursefield);
        $roomfield_l    = strtolower($roomfield);
        $subjectfield_l = strtolower($subjectfield);
        $keywordfield_l = strtolower($keywordfield);
        $startfield_l   = strtolower($startfield);
        $endfield_l     = strtolower($endfield);
        $ronidfield_l   = strtolower($ronidfield);

        $localcoursefield = $this->get_config('localcoursefield', 'id');

        $sqlfields = array($coursefield, $roomfield, $subjectfield, $keywordfield, $startfield, $endfield, $ronidfield);

        // Get appointments from external database.
        $externalappointments = array();
        $sql = $this->db_get_sql($table, array(), $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($externalappointment = $rs->FetchRow()) {
                    $externalappointment = array_change_key_case($externalappointment, CASE_LOWER);
                    // calendar events without any time information do not make sense 
                    if (!$externalappointment[$startfield_l] || !$externalappointment[$endfield_l]) {
                        continue;
                    }
                    // event end before event start does not make sense
                    if ($externalappointment[$startfield_l] > $externalappointment[$endfield_l]) {
                        continue;
                    }
                    $ronid = $externalappointment[$ronidfield_l];
                    $externalappointments[$ronid] = $externalappointment;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $extdb->Close();
            return 2;
        }

        // Close external db connection.
        $extdb->Close();

        // Find all existing course appointments with missing sitedirectory id.
        $existingappointments = [];
        $sql = "SELECT e.id, e.name, e.description, e.timestart, e.timeduration, c.$localcoursefield, sa.id AS syncid
                      FROM {event} e
                      JOIN {rwth_sd_sync_appointment} sa ON (sa.eventsource = 'rwthonline' AND e.id = sa.moodle_eventid)
                      JOIN {course} c ON c.id = e.courseid
                      WHERE e.groupid = 0";
        $rs = $DB->get_recordset_sql($sql);
        foreach ($rs as $appointment) {
            $existingappointments[] = $appointment;
        }
        $rs->close();

        // Helper function to reduce redundancy.
        $eventsareequal = function($existingevent, $externalevent) use($localcoursefield, $coursefield_l,
                $keywordfield_l, $roomfield_l, $subjectfield_l, $startfield_l, $endfield_l) {
            return $existingevent->$localcoursefield == $externalevent[$coursefield_l]
                    and $existingevent->name == $externalevent[$keywordfield_l]
                    and $existingevent->location == $externalevent[$roomfield_l]
                    and $existingevent->description == $externalevent[$subjectfield_l]
                    and $existingevent->timestart == strtotime($externalevent[$startfield_l])
                    and $existingevent->timestart + $existingevent->timeduration == strtotime($externalevent[$endfield_l]);
        };

        // Get sitedirectory ids for all synced appointments that are missing them.
        $obsoleteappointments = [];
        foreach ($existingappointments as $existingappointment) {
            $removeidx = -1;
            $updated = false;
            foreach ($externalappointments as $idx => $externalappointment) {
                if ($eventsareequal($existingappointment, $externalappointment)) {
                    $update = new stdClass();
                    $update->id = $existingappointment->syncid;
                    $update->moodle_eventid = $existingappointment->id;
                    $update->sitedirectory_eventid = $externalappointment[$ronidfield_l];
                    if (!$DB->update_record('rwth_sd_sync_appointment', $update)) {
                        $trace->output('Failed to update sitedirectory id for moodle event id '.$update->moodle_eventid.' to '.$update->sitedirectory_eventid);
                        break;
                    }
                    $trace->output('Updated sitedirectory id for moodle event id '.$update->moodle_eventid.' to '.$update->sitedirectory_eventid);
                    $updated = true;
                    $removeidx = $idx;
                    break;
                }
            }
            if (!$updated) {
                $obsoleteappointments[] = $existingappointment;
            }
            if ($removeidx != -1) {
                // For every external appointment there should only be one existing appointment.
                unset($externalappointments[$removeidx]);
            }
        }

        // we do not need these any more
        unset($externalappointments);
        unset($existingappointments);

        if ($obsoleteappointments) {
            $obsoleteeventids = [];
            require_once("$CFG->dirroot/calendar/externallib.php");

            $trace->output('Removing obsolete course appointments ...');

            // Remove sync entries.
            $transaction = $DB->start_delegated_transaction();
            $success = true;

            foreach ($obsoleteappointments as $obsoleteappointment) {
                $trace->output('Deleting event with moodle event if '.$obsoleteappointment->id.' ...');
                $success &= $DB->delete_records('rwth_sd_sync_appointment', array('moodle_eventid' => $obsoleteappointment->id));
                $obsoleteeventids[] = $obsoleteappointment->id;
            }

            if ($success) {
                // Everything done smoothly, let's commit.
                $transaction->allow_commit();
            } else {
                $errormsg = 'Error deleting records from rwth_sd_sync_appointment';
                $transaction->rollback(new moodle_exception($errormsg));
                $trace->output($errormsg);
            }
            $transaction->dispose();

            self::delete_calendar_events($obsoleteeventids);

            unset($obsoleteappointments);
        }

        $trace->output('... update of sitedirectory ids for course appointments finished.');
        $trace->finished();

        return 0;
    }

    private function add_sitedirectory_ids_for_group_appointments(progress_trace $trace) {
        global $CFG, $DB, $USER;

        if (
            !$this->get_config('dbtype')
            || !$this->get_config('groupappointmenttable')
            || !$this->get_config('groupappointmentgroupidfield')
            || !$this->get_config('groupappointmentkeywordfield')
            || !$this->get_config('groupappointmentstartfield')
            || !$this->get_config('groupappointmentendfield')
        ) {
            $syncedappointmentsexist = $DB->record_exists('rwth_sd_sync_appointment', []);
            if ($syncedappointmentsexist) {
                $trace->output('Cannot get sitedirectory ids for synced appointments because config settings are missing.');
                $trace->finished();
                return 3;
            } else {
                $trace->output('No synced appointments exist. Nothing to do.');
                $trace->finished();
                return 0;
            }
        }

        $trace->output('Starting update of sitedirectory ids for group appointments ...');

        // Init external db connection.
        if (!$extdb = $this->db_init()) {
            $trace->output('Error while communicating with external database');
            $trace->finished();
            return 1;
        }

        // We may need a lot of memory here.
        core_php_time_limit::raise();
        raise_memory_limit(MEMORY_HUGE);

        $table        = $this->get_config('groupappointmenttable');
        $groupidfield = trim($this->get_config('groupappointmentgroupidfield'));
        $roomfield    = trim($this->get_config('groupappointmentroomfield'));
        $subjectfield = trim($this->get_config('groupappointmentsubjectfield'));
        $keywordfield = trim($this->get_config('groupappointmentkeywordfield'));
        $startfield   = trim($this->get_config('groupappointmentstartfield'));
        $endfield     = trim($this->get_config('groupappointmentendfield'));
        $ronidfield   = trim($this->get_config('groupappointmentronidfield'));


        // Lowercased versions - necessary because we normalise the resultset with array_change_key_case().
        $groupidfield_l = strtolower($groupidfield);
        $roomfield_l    = strtolower($roomfield);
        $subjectfield_l = strtolower($subjectfield);
        $keywordfield_l = strtolower($keywordfield);
        $startfield_l   = strtolower($startfield);
        $endfield_l     = strtolower($endfield);
        $ronidfield_l   = strtolower($ronidfield);

        $sqlfields = [$groupidfield, $roomfield, $subjectfield, $keywordfield, $startfield, $endfield, $ronidfield];

        // Get appointments from external database.
        $externalappointments = array();
        $sql = $this->db_get_sql($table, array(), $sqlfields, true);
        if ($rs = $extdb->Execute($sql)) {
            if (!$rs->EOF) {
                while ($externalappointment = $rs->FetchRow()) {
                    $externalappointment = array_change_key_case($externalappointment, CASE_LOWER);
                    // calendar events without any time information do not make sense 
                    if (!$externalappointment[$startfield_l] || !$externalappointment[$endfield_l]) {
                        continue;
                    }
                    // event end before event start does not make sense
                    if ($externalappointment[$startfield_l] > $externalappointment[$endfield_l]) {
                        continue;
                    }
                    $ronid = $externalappointment[$ronidfield_l];
                    $externalappointments[$ronid] = $externalappointment;
                }
            }
            $rs->Close();
        } else {
            $trace->output('Error reading data from the external table');
            $extdb->Close();
            return 2;
        }

        // Close external db connection.
        $extdb->Close();

        // Find all existing course appointments with missing sitedirectory id.
        $existingappointments = [];
        $sql = "SELECT e.id, e.name, e.description, e.timestart, e.timeduration, sg.sitedirectory_groupid, sa.id AS syncid
                      FROM {event} e
                      JOIN {rwth_sd_sync_appointment} sa ON (sa.eventsource = 'rwthonline' AND e.id = sa.moodle_eventid)
                      JOIN {groups} g ON g.id = e.groupid
                      JOIN {rwth_sd_sync_group} sg ON sg.moodle_groupid = g.id";
        $rs = $DB->get_recordset_sql($sql);
        foreach ($rs as $appointment) {
            $existingappointments[] = $appointment;
        }
        $rs->close();

        // Helper function to reduce redundancy.
        $eventsareequal = function($existingevent, $externalevent) use($groupidfield_l, $keywordfield_l,
        $roomfield_l, $subjectfield_l, $startfield_l, $endfield_l) {
            return $existingevent->sitedirectory_groupid == $externalevent[$groupidfield_l]
                    and $existingevent->name == $externalevent[$keywordfield_l]
                    and $existingevent->location == $externalevent[$roomfield_l]
                    and $existingevent->description == $externalevent[$subjectfield_l]
                    and $existingevent->timestart == strtotime($externalevent[$startfield_l])
                    and $existingevent->timestart + $existingevent->timeduration == strtotime($externalevent[$endfield_l]);
        };

        // Get sitedirectory ids for all synced appointments that are missing them.
        $obsoleteappointments = [];
        foreach ($existingappointments as $existingappointment) {
            $removeidx = -1;
            $updated = false;
            foreach ($externalappointments as $idx => $externalappointment) {
                if ($eventsareequal($existingappointment, $externalappointment)) {
                    $update = new stdClass();
                    $update->id = $existingappointment->syncid;
                    $update->moodle_eventid = $existingappointment->id;
                    $update->sitedirectory_eventid = $externalappointment[$ronidfield_l];
                    if (!$DB->update_record('rwth_sd_sync_appointment', $update)) {
                        $trace->output('Failed to update sitedirectory id for moodle event id '.$update->moodle_eventid.' to '.$update->sitedirectory_eventid);
                        break;
                    }
                    $trace->output('Updated sitedirectory id for moodle event id '.$update->moodle_eventid.' to '.$update->sitedirectory_eventid);
                    $updated = true;
                    $removeidx = $idx;
                    break;
                }
            }
            if (!$updated) {
                $obsoleteappointments[] = $existingappointment;
            }
            if ($removeidx != -1) {
                // For every external appointment there should only be one existing appointment.
                unset($externalappointments[$removeidx]);
            }
        }

        // we do not need these any more
        unset($externalappointments);
        unset($existingappointments);

        if ($obsoleteappointments) {
            $obsoleteeventids = [];
            require_once("$CFG->dirroot/calendar/externallib.php");

            $trace->output('Removing obsolete group appointments ...');

            // Remove sync entries.
            $transaction = $DB->start_delegated_transaction();
            $success = true;

            foreach ($obsoleteappointments as $obsoleteappointment) {
                $trace->output('Deleting event with moodle event id '.$obsoleteappointment->id.' ...');
                $success &= $DB->delete_records('rwth_sd_sync_appointment', array('moodle_eventid' => $obsoleteappointment->id));
                $obsoleteeventids[] = $obsoleteappointment->id;
            }

            if ($success) {
                // Everything done smoothly, let's commit.
                $transaction->allow_commit();
            } else {
                $errormsg = 'Error deleting records from rwth_sd_sync_appointment';
                $transaction->rollback(new moodle_exception($errormsg));
                $trace->output($errormsg);
            }
            $transaction->dispose();

            self::delete_calendar_events($obsoleteeventids);

            unset($obsoleteappointments);
        }

        $trace->output('... update of sitedirectory ids for group appointments finished.');
        $trace->finished();

        return 0;
    }

    /** upgrade functions end *************************************************/

    protected function db_get_sql($table, array $conditions, array $fields, $distinct = false, $sort = "") {
        $fields = $fields ? implode(',', $fields) : "*";
        $where = array();
        if ($conditions) {
            foreach ($conditions as $key=>$value) {
                $value = $this->db_encode($this->db_addslashes($value));

                $where[] = "$key = '$value'";
            }
        }
        $where = $where ? "WHERE ".implode(" AND ", $where) : "";
        $sort = $sort ? "ORDER BY $sort" : "";
        $distinct = $distinct ? "DISTINCT" : "";
        $sql = "SELECT $distinct $fields
                  FROM $table
                 $where
                  $sort";

        return $sql;
    }

    /**
     * Tries to make connection to the external database.
     *
     * @return null|ADONewConnection
     */
    protected function db_init() {
        global $CFG;

        require_once($CFG->libdir.'/adodb/adodb.inc.php');

        // Connect to the external database (forcing new connection).
        $extdb = ADONewConnection($this->get_config('dbtype'));
        if ($this->get_config('debugdb')) {
            $extdb->debug = true;
            ob_start(); // Start output buffer to allow later use of the page headers.
        }

        // The dbtype my contain the new connection URL, so make sure we are not connected yet.
        if (!$extdb->IsConnected()) {
            $result = $extdb->Connect($this->get_config('dbhost'), $this->get_config('dbuser'), $this->get_config('dbpass'), $this->get_config('dbname'), true);
            if (!$result) {
                return null;
            }
        }

        $extdb->SetFetchMode(ADODB_FETCH_ASSOC);
        if ($this->get_config('dbsetupsql')) {
            $extdb->Execute($this->get_config('dbsetupsql'));
        }
        return $extdb;
    }

    protected function db_addslashes($text) {
        // Use custom made function for now - it is better to not rely on adodb or php defaults.
        if ($this->get_config('dbsybasequoting')) {
            $text = str_replace('\\', '\\\\', $text);
            $text = str_replace(array('\'', '"', "\0"), array('\\\'', '\\"', '\\0'), $text);
        } else {
            $text = str_replace("'", "''", $text);
        }
        return $text;
    }

    protected function db_encode($text) {
        $dbenc = $this->get_config('dbencoding');
        if (empty($dbenc) or $dbenc == 'utf-8') {
            return $text;
        }
        if (is_array($text)) {
            foreach ($text as $k => $value) {
                $text[$k] = $this->db_encode($value);
            }
            return $text;
        } else {
            return \core_text::convert($text, 'utf-8', $dbenc);
        }
    }

    protected function db_decode($text) {
        $dbenc = $this->get_config('dbencoding');
        if (empty($dbenc) or $dbenc == 'utf-8') {
            return $text;
        }
        if (is_array($text)) {
            foreach ($text as $k => $value) {
                $text[$k] = $this->db_decode($value);
            }
            return $text;
        } else {
            return \core_text::convert($text, $dbenc, 'utf-8');
        }
    }

    /**
     * Test plugin settings, print info to output.
     */
    public function test_settings() {
        global $CFG, $OUTPUT;

        // NOTE: this is not localised intentionally, admins are supposed to understand English at least a bit...

        raise_memory_limit(MEMORY_HUGE);

        $this->load_config();

        $coursetable = $this->get_config('newcoursetable');

        if (empty($coursetable)) {
            echo $OUTPUT->notification('External course table not specified.', 'notifyproblem');
        }

        $olddebug = $CFG->debug;
        $olddisplay = ini_get('display_errors');
        ini_set('display_errors', '1');
        $CFG->debug = DEBUG_DEVELOPER;
        $olddebugdb = $this->config->debugdb;
        $this->config->debugdb = 1;
        error_reporting($CFG->debug);

        $adodb = $this->db_init();

        if (!$adodb or !$adodb->IsConnected()) {
            $this->config->debugdb = $olddebugdb;
            $CFG->debug = $olddebug;
            ini_set('display_errors', $olddisplay);
            error_reporting($CFG->debug);
            ob_end_flush();

            echo $OUTPUT->notification('Cannot connect the database.', 'notifyproblem');
            return;
        }

        if (!empty($coursetable)) {
            $rs = $adodb->Execute("SELECT *
                                     FROM $coursetable");
            if (!$rs) {
                echo $OUTPUT->notification('Can not read external course table.', 'notifyproblem');

            } else if ($rs->EOF) {
                echo $OUTPUT->notification('External course table is empty.', 'notifyproblem');
                $rs->Close();

            } else {
                $fields_obj = $rs->FetchObj();
                $columns = array_keys((array)$fields_obj);

                echo $OUTPUT->notification('External course table contains following columns:<br />'.implode(', ', $columns), 'notifysuccess');
                $rs->Close();
            }
        }

        $adodb->Close();

        $this->config->debugdb = $olddebugdb;
        $CFG->debug = $olddebug;
        ini_set('display_errors', $olddisplay);
        error_reporting($CFG->debug);
        ob_end_flush();
    }

    /** from enrollib baseclass : **********************************************/

    /**
     * Returns name of this enrol plugin
     * @return string
     */
    public function get_name() {
        return 'rwth_sd_sync';
    }

    /**
     * Makes sure config is loaded and cached.
     * @return void
     */
    protected function load_config() {
        if (!isset($this->config)) {
            $name = $this->get_name();
            $this->config = get_config("local_$name");
        }
    }

    /**
     * Returns plugin config value
     * @param  string $name
     * @param  string $default value if config does not exist yet
     * @return string value or default
     */
    public function get_config($name, $default = NULL) {
        $this->load_config();
        return isset($this->config->$name) ? $this->config->$name : $default;
    }

    /**
     * Sets plugin config value
     * @param  string $name name of config
     * @param  string $value string config value, null means delete
     * @return string value
     */
    public function set_config($name, $value) {
        $pluginname = $this->get_name();
        $this->load_config();
        if ($value === NULL) {
            unset($this->config->$name);
        } else {
            $this->config->$name = $value;
        }
        set_config($name, $value, "local_$pluginname");
    }

    protected static function delete_calendar_events(array $eventids) {
        global $DB, $USER;
        // make me admin
        // @TODO: better to create and get generic manager for each course!
        $userid = $USER->id;
        self::setUser(2);
        try {
            $transaction = $DB->start_delegated_transaction();
            foreach ($eventids as $eventid) {
                $eventobj = calendar_event::load($eventid);
                $eventobj->delete(false);
            }
            $transaction->allow_commit();
        } finally {
            self::setUser($userid);
        }
    }
}
