<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

defined('MOODLE_INTERNAL') || die;

use local_rwth_sd_sync\sdapi;

function xmldb_local_rwth_sd_sync_upgrade($oldversion) {
    global $DB;
    $dbman = $DB->get_manager();

    // Rename appointment table.
    if ($oldversion < 2019021500) {
        $oldname = 'local_rwth_sd_sync';
        $newname = 'rwth_sd_sync_appointment';
        if ($dbman->table_exists($oldname)) {
            $oldtable = new xmldb_table($oldname);
            if ($dbman->table_exists($newname)) {
                $dbman->drop_table($oldtable);
            } else {
                $dbman->rename_table($oldtable, $newname);
            }
        }
        upgrade_plugin_savepoint(true, 2019021500, 'local', 'rwth_sd_sync');
    }

    // Fix foreign key for appointments table.
    if ($oldversion < 2019021501) {
        $table = new xmldb_table('rwth_sd_sync_appointment');
        $oldkey = new xmldb_key('eventid-id', XMLDB_KEY_FOREIGN, ['eventid'], 'mdl_event', ['id']);
        $newkey = new xmldb_key('fk-event', XMLDB_KEY_FOREIGN, ['eventid'], 'event', ['id']);
        $dbman->drop_key($table, $oldkey);
        $dbman->add_key($table, $newkey);
        upgrade_plugin_savepoint(true, 2019021501, 'local', 'rwth_sd_sync');
    }

    // Add group table.
    if ($oldversion < 2019021900) {
        $table = new xmldb_table('rwth_sd_sync_group');
        $table->add_field('id', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, XMLDB_SEQUENCE, null);
        $table->add_field('sitedirectory_groupid', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, null, null);
        $table->add_field('moodle_groupid', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, null, null);
        $table->add_key('primary', XMLDB_KEY_PRIMARY, ['id']);
        $table->add_key('un-sdgroupid', XMLDB_KEY_UNIQUE, ['sitedirectory_groupid']);
        $table->add_key('fk-course', XMLDB_KEY_FOREIGN_UNIQUE, ['moodle_groupid'], 'group', ['id']);
        if (!$dbman->table_exists($table)) {
            $dbman->create_table($table);
        }
        upgrade_plugin_savepoint(true, 2019021900, 'local', 'rwth_sd_sync');
    }

    if ($oldversion < 2019022104) {
        $sd = new sdapi();
        $trace = new html_progress_trace();
        $sd->write_courseids_to_sitedirectory($trace);
    }

    // Add group_participant table.
    if ($oldversion < 2019042900) {
        $table = new xmldb_table('rwth_sd_sync_group_member');
        $table->add_field('id', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, XMLDB_SEQUENCE, null);
        $table->add_field('moodle_groupid', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, null, null);
        $table->add_field('moodle_userid', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, null, null);
        $table->add_key('primary', XMLDB_KEY_PRIMARY, ['id']);
        $table->add_key('fk-group', XMLDB_KEY_FOREIGN_UNIQUE, ['moodle_groupid'], 'group', ['id']);
        $table->add_key('fk-user', XMLDB_KEY_FOREIGN_UNIQUE, ['moodle_userid'], 'user', ['id']);
        if (!$dbman->table_exists($table)) {
            $dbman->create_table($table);
        }
        upgrade_plugin_savepoint(true, 2019042900, 'local', 'rwth_sd_sync');
    }

    // Fix foreign keys for group participants table.
    if ($oldversion < 2019043000) {
        $table = new xmldb_table('rwth_sd_sync_group_member');
        $oldkey = new xmldb_key('fk-group', XMLDB_KEY_FOREIGN_UNIQUE, ['moodle_groupid'], 'group', ['id']);
        $newkey = new xmldb_key('fk-group', XMLDB_KEY_FOREIGN, ['moodle_groupid'], 'group', ['id']);
        $dbman->drop_key($table, $oldkey);
        $dbman->add_key($table, $newkey);
        $oldkey = new xmldb_key('fk-user', XMLDB_KEY_FOREIGN_UNIQUE, ['moodle_userid'], 'user', ['id']);
        $newkey = new xmldb_key('fk-user', XMLDB_KEY_FOREIGN, ['moodle_userid'], 'user', ['id']);
        $dbman->drop_key($table, $oldkey);
        $dbman->add_key($table, $newkey);
        upgrade_plugin_savepoint(true, 2019043000, 'local', 'rwth_sd_sync');
    }

    // Add sitedirectory ID for appointments (at first nullable).
    if ($oldversion < 2019043001) {
        $table = new xmldb_table('rwth_sd_sync_appointment');
        $oldfield = new xmldb_field('eventid', XMLDB_TYPE_INTEGER, 10, null, XMLDB_NOTNULL, null, null);
        $dbman->rename_field($table, $oldfield, 'moodle_eventid');
        $newfield = new xmldb_field('sitedirectory_eventid', XMLDB_TYPE_INTEGER, 10, null, null, null, null);
        $dbman->add_field($table, $newfield);
        upgrade_plugin_savepoint(true, 2019043001, 'local', 'rwth_sd_sync');
    }

    // Add sitedirectory ID for existing appointments.
    if ($oldversion < 2019043003) {
        $trace = new \text_progress_trace();
        $syncer = new \local_rwth_sd_sync\sdapi();
        $ret = $syncer->add_sitedirectory_ids_for_appointments($trace);
        if ($ret != 0) {
            return false;
        }
        upgrade_plugin_savepoint(true, 2019043003, 'local', 'rwth_sd_sync');
    }

    // Enable unique constraint for sitedirectory ID.
    if ($oldversion < 2019043006) {
        $table = new xmldb_table('rwth_sd_sync_appointment');
        $key = new xmldb_key('un-sdeventid', XMLDB_KEY_UNIQUE, ['sitedirectory_eventid']);
        $dbman->add_key($table, $key);
        upgrade_plugin_savepoint(true, 2019043006, 'local', 'rwth_sd_sync');
    }

    if ($oldversion < 2019060300) {
        // Delete all synced events and respective sync entries.
        $DB->execute('DELETE e.*
            FROM {event} e
            JOIN {rwth_sd_sync_appointment} sa
                ON e.id = sa.moodle_eventid');
        $DB->delete_records('rwth_sd_sync_appointment');

        // Make sitedirectory_eventid column not nullable.
        $table = new xmldb_table('rwth_sd_sync_appointment');
        $key = new xmldb_key('un-sdeventid', XMLDB_KEY_UNIQUE, ['sitedirectory_eventid']);
        $dbman->drop_key($table, $key);

        $oldfield = new xmldb_field('sitedirectory_eventid', XMLDB_TYPE_INTEGER, 10, null, null, null, null);
        $newfield = new xmldb_field('sitedirectory_eventid', XMLDB_TYPE_INTEGER, 10, null, XMLDB_NOTNULL, null, null);
        $dbman->drop_field($table, $oldfield);
        $dbman->add_field($table, $newfield);

        $dbman->add_key($table, $key);

        // Run an appointment sync to add back all the appointments again.
        $sd = new sdapi();
        $trace = new html_progress_trace();
        $sd->sync_appointments($trace);

        upgrade_plugin_savepoint(true, 2019060300, 'local', 'rwth_sd_sync');
    }

    return true;
}
